@@ -1,7 +1,11 @@
 #include <glib/gtypes.h>
 
 double
-rsvg_css_parse_length (const char *str, int *fixed);
+rsvg_css_parse_length (const char *str, gdouble pixels_per_inch, int *fixed);
+
+double
+rsvg_css_parse_normalized_length(const char *str, gdouble pixels_per_inch,
+				 gdouble normalize_to);
 
 gboolean
 rsvg_css_param_match (const char *str, const char *param_name);
