@@ -2,6 +2,7 @@
    rsvg.c: SAX-based renderer for SVG files into a GdkPixbuf.
 
    Copyright (C) 2000 Eazel, Inc.
+   Copyright (C) 2002 Dom Lachowicz <cinamod@hotmail.com>
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public License as
@@ -51,6 +52,14 @@
 
 #define SVG_BUFFER_SIZE (1024 * 8)
 
+/* 4/3 * (1-cos 45)/sin 45 = 4/3 * sqrt(2) - 1 */
+#define RSVG_ARC_MAGIC ((double) 0.5522847498)
+
+/*
+ * Ideally this will be configurable
+ */
+#define RSVG_PIXELS_PER_INCH 72.0
+
 typedef struct {
   double affine[6];
 
@@ -174,19 +183,32 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
   art_u8 *pixels;
   gint fixed;
   RsvgState *state;
-  gboolean has_alpha = 1;
+  gboolean has_alpha = TRUE;
   gint new_width, new_height;
-  double x_zoom;
-  double y_zoom;
+  double x_zoom = 1.;
+  double y_zoom = 1.;
+
+  double vbox_x = 0, vbox_y = 0, vbox_w = 0, vbox_h = 0;
+  gboolean has_vbox = TRUE;
 
   if (atts != NULL)
     {
       for (i = 0; atts[i] != NULL; i += 2)
 	{
+	  /* x & y should be ignored since we should always be the outermost SVG,
+	     at least for now */
 	  if (!strcmp ((char *)atts[i], "width"))
-	    width = rsvg_css_parse_length ((char *)atts[i + 1], &fixed);
-	  else if (!strcmp ((char *)atts[i], "height"))
-	    height = rsvg_css_parse_length ((char *)atts[i + 1], &fixed);
+	    width = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "height"))	    
+	    height = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "viewBox"))
+	    {
+	      /* todo: viewbox can have whitespace and/or comma but we're only likely to see
+		 these 2 combinations */
+	      if (4 == sscanf ((char *)atts[i + 1], " %lf %lf %lf %lf ", &vbox_x, &vbox_y, &vbox_w, &vbox_h) ||
+		  4 == sscanf ((char *)atts[i + 1], " %lf , %lf , %lf , %lf ", &vbox_x, &vbox_y, &vbox_w, &vbox_h))
+		has_vbox = TRUE;
+	    }
 	}
 
 #ifdef VERBOSE
@@ -194,17 +216,34 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
 	       width, height);
 #endif
 
-      if (width == 0 || height == 0)
+      if (width <= 0 || height <= 0)
         {
           /* FIXME: GError here? */
           g_warning ("rsvg_start_svg: can't render 0-sized SVG");
           return;
         }
 
-      new_width = width;
-      new_height = height;
-      if (ctx->size_func)
-	(* ctx->size_func) (&new_width, &new_height, ctx->user_data);
+      if (has_vbox && vbox_w > 0. && vbox_h > 0.)
+	{
+	  new_width  = (int)floor (vbox_w - vbox_x);
+	  new_height = (int)floor (vbox_h - vbox_y);
+
+	  /* apply the sizing function on the *original* width and height
+	     to acquire our real destination size. we'll scale it against
+	     the viewBox's coordinates later */
+	  if (ctx->size_func)
+	    (* ctx->size_func) (&height, &height, ctx->user_data);
+	}
+      else
+	{
+	  new_width  = width;
+	  new_height = height;
+
+	  /* apply the sizing function to acquire our new width and height.
+	     we'll scale this against the old values later */
+	  if (ctx->size_func)
+	    (* ctx->size_func) (&new_width, &new_height, ctx->user_data);
+	}
 
       if (new_width == 0 || new_height == 0)
 	{
@@ -213,8 +252,20 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
           return;
 	}
 
-      x_zoom = (width < 0 || new_width < 0) ? 1 : (double) new_width / width;
-      y_zoom = (height < 0 || new_height < 0) ? 1 : (double) new_height / height;
+      if (!has_vbox)
+	{
+	  x_zoom = (width < 0 || new_width < 0) ? 1 : (double) new_width / width;
+	  y_zoom = (height < 0 || new_height < 0) ? 1 : (double) new_height / height;
+	}
+      else
+	{	  
+	  x_zoom = (width < 0 || new_width < 0) ? 1 : (double) width / new_width;
+	  y_zoom = (height < 0 || new_height < 0) ? 1 : (double) height / new_height;
+
+	  /* reset these so that we get a properly sized SVG and not a huge one */
+	  new_width  = width;
+	  new_height = height;
+	}
 
       /* Scale size of target pixbuf */
       state = &ctx->state[ctx->n_state - 1];
@@ -289,7 +340,8 @@ rsvg_parse_style_arg (RsvgHandle *ctx, RsvgState *state, const char *str)
   else if (rsvg_css_param_match (str, "stroke-width"))
     {
       int fixed;
-      state->stroke_width = rsvg_css_parse_length (str + arg_off, &fixed);
+      /* should be rsvg_css_parse_normalized_length */
+      state->stroke_width = rsvg_css_parse_length (str + arg_off, RSVG_PIXELS_PER_INCH, &fixed);
     }
   else if (rsvg_css_param_match (str, "stroke-linecap"))
     {
@@ -336,6 +388,46 @@ rsvg_parse_style_arg (RsvgHandle *ctx, RsvgState *state, const char *str)
     }
 }
 
+/* tell whether @str is a supported style argument 
+   whenever something gets added to parse_arg, please
+   remember to add it here too
+*/
+static gboolean
+rsvg_is_style_arg(const char *str)
+{
+  static GHashTable *styles = NULL;
+  if (!styles)
+    {
+      styles = g_hash_table_new (g_str_hash, g_str_equal);
+      
+      g_hash_table_insert (styles, "opacity",         GINT_TO_POINTER (TRUE));
+      g_hash_table_insert (styles, "fill",            GINT_TO_POINTER (TRUE));
+      g_hash_table_insert (styles, "fill-opacity",    GINT_TO_POINTER (TRUE));
+      g_hash_table_insert (styles, "stroke",          GINT_TO_POINTER (TRUE));
+      g_hash_table_insert (styles, "stroke-width",    GINT_TO_POINTER (TRUE));
+      g_hash_table_insert (styles, "stroke-linecap",  GINT_TO_POINTER (TRUE));
+      g_hash_table_insert (styles, "stroke-opacity",  GINT_TO_POINTER (TRUE));
+      g_hash_table_insert (styles, "stroke-linejoin", GINT_TO_POINTER (TRUE));
+      g_hash_table_insert (styles, "font-size",       GINT_TO_POINTER (TRUE));
+      g_hash_table_insert (styles, "font-family",     GINT_TO_POINTER (TRUE));
+      g_hash_table_insert (styles, "stop-color",      GINT_TO_POINTER (TRUE));
+      g_hash_table_insert (styles, "stop-opacity",    GINT_TO_POINTER (TRUE));
+    }
+  
+  /* this will default to 0 (FALSE) on a failed lookup */
+  return GPOINTER_TO_INT (g_hash_table_lookup (styles, str)); 
+}
+
+/* take a pair of the form (fill="#ff00ff") and parse it as a style */
+static void
+rsvg_parse_style_pair (RsvgHandle *ctx, RsvgState *state, 
+		       const char *key, const char *val)
+{
+  gchar * str = g_strdup_printf ("%s:%s", key, val);
+  rsvg_parse_style_arg (ctx, state, str);
+  g_free (str);
+}
+
 /* Split a CSS2 style into individual style arguments, setting attributes
    in the SVG context.
 
@@ -421,7 +513,7 @@ rsvg_parse_transform (double dst[6], const char *src)
 	    {
 	      if (n_args == sizeof(args) / sizeof(args[0]))
 		return FALSE; /* too many args */
-	      args[n_args] = strtod (src + idx, &end_ptr);
+	      args[n_args] = g_ascii_strtod (src + idx, &end_ptr);
 	      idx = end_ptr - src;
 
 	      while (g_ascii_isspace (src[idx]))
@@ -539,6 +631,9 @@ rsvg_parse_style_attrs (RsvgHandle *ctx, const xmlChar **atts)
 	  else if (!strcmp ((char *)atts[i], "transform"))
 	    rsvg_parse_transform_attr (ctx, &ctx->state[ctx->n_state - 1],
 				       (char *)atts[i + 1]);
+	  else if (rsvg_is_style_arg ((char *)atts[i]))
+	    rsvg_parse_style_pair (ctx, &ctx->state[ctx->n_state - 1],
+				   (char *)atts[i], (char *)atts[i + 1]);
 	}
     }
 }
@@ -885,6 +980,7 @@ rsvg_start_path (RsvgHandle *ctx, const xmlChar **atts)
 	    d = (char *)atts[i + 1];
 	}
     }
+
   if (d != NULL)
     {
       RsvgBpathDef *bpath_def;
@@ -905,8 +1001,6 @@ typedef struct _RsvgSaxHandlerText RsvgSaxHandlerText;
 struct _RsvgSaxHandlerText {
   RsvgSaxHandler super;
   RsvgHandle *ctx;
-  double xpos;
-  double ypos;
 };
 
 static void
@@ -964,7 +1058,7 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
 #endif
 
   if (ctx->pango_context == NULL)
-    ctx->pango_context = pango_ft2_get_context (72, 72); /* FIXME: dpi? */
+    ctx->pango_context = pango_ft2_get_context ((guint)RSVG_PIXELS_PER_INCH, (guint)RSVG_PIXELS_PER_INCH); /* FIXME: dpi? */
 
   has_alpha = gdk_pixbuf_get_has_alpha (pixbuf);
   
@@ -984,7 +1078,10 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
   font = pango_font_description_copy (pango_context_get_font_description (ctx->pango_context));
   if (state->font_family)
     pango_font_description_set_family_static (font, state->font_family);
-  pango_font_description_set_size (font, state->font_size * PANGO_SCALE);
+
+  /* we need to resize the font by our X or Y scale (ideally could stretch in both directions...)
+     which, though? Y for now */
+  pango_font_description_set_size (font, state->font_size * PANGO_SCALE * state->affine[3]);
   pango_layout_set_font_description (layout, font);
   pango_font_description_free (font);
   
@@ -1026,18 +1123,123 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
   g_free (string);
 }
 
+static void
+rsvg_start_tspan (RsvgHandle *ctx, const xmlChar **atts)
+{
+  int i, fixed;
+  double affine[6] ;
+  double x, y, dx, dy;
+  RsvgState *state;
+  x = y = dx = dy = 0.;
+
+  if (atts != NULL)
+    {
+      for (i = 0; atts[i] != NULL; i += 2)
+	{
+	  if (!strcmp ((char *)atts[i], "x"))
+	    x = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  if (!strcmp ((char *)atts[i], "y"))
+	    y = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  if (!strcmp ((char *)atts[i], "dx"))
+	    dx = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  if (!strcmp ((char *)atts[i], "dy"))
+	    dy = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	}
+    }
+
+  state = &ctx->state[ctx->n_state - 1];
+
+  x += dx ;
+  y += dy ;
+
+  if (x >= 0 && y >= 0)
+    {
+      art_affine_translate (affine, x, y);
+      art_affine_multiply (state->affine, affine, state->affine);
+    }
+
+  /* todo: transform() is illegal here */
+  rsvg_parse_style_attrs (ctx, atts);
+
+#ifdef VERBOSE
+  fprintf (stderr, "begin tspan!\n");
+#endif
+}
+
+static void
+rsvg_text_handler_start (RsvgSaxHandler *self, const xmlChar *name,
+			 const xmlChar **atts)
+{
+  RsvgSaxHandlerText *z = (RsvgSaxHandlerText *)self;
+  RsvgHandle *ctx = z->ctx;
+
+  /* push the state stack */
+  if (ctx->n_state == ctx->n_state_max)
+    ctx->state = g_renew (RsvgState, ctx->state, ctx->n_state_max <<= 1);
+  if (ctx->n_state)
+    rsvg_state_clone (&ctx->state[ctx->n_state],
+		      &ctx->state[ctx->n_state - 1]);
+  else
+    rsvg_state_init (ctx->state);
+  ctx->n_state++;
+  
+  if (!strcmp ((char *)name, "tspan"))
+    rsvg_start_tspan (ctx, atts);
+}
+
+static void
+rsvg_text_handler_end (RsvgSaxHandler *self, const xmlChar *name)
+{
+  RsvgSaxHandlerText *z = (RsvgSaxHandlerText *)self;
+  RsvgHandle *ctx = z->ctx;
+
+  /* pop the state stack */
+  ctx->n_state--;
+  rsvg_state_finalize (&ctx->state[ctx->n_state]);
+}
+
 static void
 rsvg_start_text (RsvgHandle *ctx, const xmlChar **atts)
 {
-  RsvgSaxHandlerText *handler = g_new0 (RsvgSaxHandlerText, 1);
+  int i, fixed;
+  double affine[6] ;
+  double x, y, dx, dy;
+  RsvgState *state;
 
+  RsvgSaxHandlerText *handler = g_new0 (RsvgSaxHandlerText, 1);
   handler->super.free = rsvg_text_handler_free;
   handler->super.characters = rsvg_text_handler_characters;
+  handler->super.start_element = rsvg_text_handler_start;
+  handler->super.end_element   = rsvg_text_handler_end;
   handler->ctx = ctx;
 
-  /* todo: parse "x" and "y" attributes */
-  handler->xpos = 0;
-  handler->ypos = 0;
+  x = y = dx = dy = 0.;
+
+  if (atts != NULL)
+    {
+      for (i = 0; atts[i] != NULL; i += 2)
+	{
+	  if (!strcmp ((char *)atts[i], "x"))
+	    x = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  if (!strcmp ((char *)atts[i], "y"))
+	    y = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  if (!strcmp ((char *)atts[i], "dx"))
+	    dx = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  if (!strcmp ((char *)atts[i], "dy"))
+	    dy = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	}
+    }
+
+  state = &ctx->state[ctx->n_state - 1];
+
+  x += dx ;
+  y += dy ;
+
+  if (x >= 0 && y >= 0)
+    {
+      art_affine_translate (affine, x, y);
+      art_affine_multiply (state->affine, affine, state->affine);
+    }
 
   rsvg_parse_style_attrs (ctx, atts);
   ctx->handler = &handler->super;
@@ -1079,7 +1281,6 @@ rsvg_gradient_stop_handler_start (RsvgSaxHandler *self, const xmlChar *name,
   int i;
   double offset = 0;
   gboolean got_offset = FALSE;
-  gint fixed;
   RsvgState state;
   int n_stop;
 
@@ -1097,7 +1298,7 @@ rsvg_gradient_stop_handler_start (RsvgSaxHandler *self, const xmlChar *name,
 	{
 	  if (!strcmp ((char *)atts[i], "offset"))
 	    {
-	      offset = rsvg_css_parse_length ((char *)atts[i + 1], &fixed);
+	      offset = g_ascii_strtod ((char *)atts[i + 1], NULL);
 	      got_offset = TRUE;
 	    }
 	  else if (!strcmp ((char *)atts[i], "style"))
@@ -1166,6 +1367,7 @@ rsvg_start_linear_gradient (RsvgHandle *ctx, const xmlChar **atts)
   char *id = NULL;
   double x1 = 0, y1 = 0, x2 = 100, y2 = 0;
   ArtGradientSpread spread = ART_GRADIENT_PAD;
+  gint fixed = FALSE;
 
   /* todo: only handles numeric coordinates in gradientUnits = userSpace */
   if (atts != NULL)
@@ -1175,13 +1377,13 @@ rsvg_start_linear_gradient (RsvgHandle *ctx, const xmlChar **atts)
 	  if (!strcmp ((char *)atts[i], "id"))
 	    id = (char *)atts[i + 1];
 	  else if (!strcmp ((char *)atts[i], "x1"))
-	    x1 = atof ((char *)atts[i + 1]);
+	    x1 = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
 	  else if (!strcmp ((char *)atts[i], "y1"))
-	    y1 = atof ((char *)atts[i + 1]);
+	    y1 = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
 	  else if (!strcmp ((char *)atts[i], "x2"))
-	    x2 = atof ((char *)atts[i + 1]);
+	    x2 = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
 	  else if (!strcmp ((char *)atts[i], "y2"))
-	    y2 = atof ((char *)atts[i + 1]);
+	    y2 = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
 	  else if (!strcmp ((char *)atts[i], "spreadMethod"))
 	    {
 	      if (!strcmp ((char *)atts[i + 1], "pad"))
@@ -1229,6 +1431,7 @@ rsvg_start_radial_gradient (RsvgHandle *ctx, const xmlChar **atts)
   int i;
   char *id = NULL;
   double cx = 50, cy = 50, r = 50, fx = 50, fy = 50;
+  gint fixed = FALSE;
 
   /* todo: only handles numeric coordinates in gradientUnits = userSpace */
   if (atts != NULL)
@@ -1238,15 +1441,15 @@ rsvg_start_radial_gradient (RsvgHandle *ctx, const xmlChar **atts)
 	  if (!strcmp ((char *)atts[i], "id"))
 	    id = (char *)atts[i + 1];
 	  else if (!strcmp ((char *)atts[i], "cx"))
-	    cx = atof ((char *)atts[i + 1]);
+	    cx = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
 	  else if (!strcmp ((char *)atts[i], "cy"))
-	    cy = atof ((char *)atts[i + 1]);
+	    cy = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
 	  else if (!strcmp ((char *)atts[i], "r"))
-	    r = atof ((char *)atts[i + 1]);
+	    r = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
 	  else if (!strcmp ((char *)atts[i], "fx"))
-	    fx = atof ((char *)atts[i + 1]);
+	    fx = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
 	  else if (!strcmp ((char *)atts[i], "fy"))
-	    fy = atof ((char *)atts[i + 1]);
+	    fy = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
 	}
     }
 
@@ -1267,6 +1470,364 @@ rsvg_start_radial_gradient (RsvgHandle *ctx, const xmlChar **atts)
   grad->fy = fy;
 }
 
+static GString *
+rsvg_make_poly_point_list(const char * points)
+{
+  guint idx = 0, size = strlen(points);
+  GString * str = g_string_sized_new (size);
+	
+  while (idx < size) 
+    {
+      /* scan for first point */
+      while (!g_ascii_isdigit (points[idx]) && (points[idx] != '.') 
+	     && (points[idx] != '-') && (idx < size))
+	idx++;
+      
+      /* now build up the point list (everything until next letter!) */
+      if (idx < size && points[idx] == '-')
+	g_string_append_c (str, points[idx++]); /* handle leading '-' */
+      while ((g_ascii_isdigit (points[idx]) || (points[idx] == '.')) && (idx < size)) 
+	  g_string_append_c (str, points[idx++]);
+
+      g_string_append_c (str, ' ');
+    }
+
+  return str;
+}
+
+static void
+rsvg_start_any_poly(RsvgHandle *ctx, const xmlChar **atts, gboolean is_polyline)
+{
+  /* the only difference i'm making between polygon and polyline is
+     that a polyline closes the path */
+
+  int i;
+  const char * verts = (const char *)NULL;
+  GString * g = NULL;
+  gchar ** pointlist = NULL;
+  xmlChar *path_atts[3];
+
+  rsvg_parse_style_attrs (ctx, atts);
+  if (atts != NULL)
+    {
+      for (i = 0; atts[i] != NULL; i += 2)
+	{
+	  /* support for svg < 1.0 which used verts */
+	  if (!strcmp ((char *)atts[i], "verts") || !strcmp ((char *)atts[i], "points"))
+	    verts = (const char *)atts[i + 1];
+	}
+    }
+
+  if (!verts)
+    return;
+
+  /* todo: make the following more memory and CPU friendly */
+  g = rsvg_make_poly_point_list (verts);
+  pointlist = g_strsplit (g->str, " ", -1);
+  g_string_free (g, TRUE);
+
+  /* represent as a "moveto, lineto*, close" path */  
+  if (pointlist)
+    {
+      GString * d = g_string_sized_new (strlen(verts));
+      g_string_append_printf (d, "M %s %s ", pointlist[0], pointlist[1] );
+
+      for (i = 2; pointlist[i] != NULL && pointlist[i][0] != '\0'; i += 2)
+	  g_string_append_printf (d, "L %s %s ", pointlist[i], pointlist[i+1]);
+
+      if (!is_polyline)
+	g_string_append (d, "Z");
+
+      path_atts[0] = (xmlChar*)"d";
+      path_atts[1] = (xmlChar*)d->str;
+      path_atts[2] = (xmlChar*)NULL;
+      rsvg_start_path (ctx, (const xmlChar **)path_atts);
+
+      g_string_free (d, TRUE);
+      g_strfreev(pointlist);
+    }
+}
+
+static void
+rsvg_start_polygon (RsvgHandle *ctx, const xmlChar **atts)
+{
+  rsvg_start_any_poly (ctx, atts, FALSE);
+}
+
+static void
+rsvg_start_polyline (RsvgHandle *ctx, const xmlChar **atts)
+{
+  rsvg_start_any_poly (ctx, atts, TRUE);
+}
+
+static void
+rsvg_start_image (RsvgHandle *ctx, const xmlChar **atts)
+{
+  int i;
+  double x = 0, y = 0, w = 0, h = 0;
+  const char * href = (const char *)NULL;
+  GdkPixbuf *pixbuf, *img;
+  gboolean has_alpha;
+
+  rsvg_parse_style_attrs (ctx, atts);
+  if (atts != NULL)
+    {
+      gint fixed = FALSE;
+      
+      for (i = 0; atts[i] != NULL; i += 2)
+	{
+	  if (!strcmp ((char *)atts[i], "x"))
+	    x = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "y"))
+	    y = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "width"))
+	    w = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "height"))
+	    h = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  /* path is used by some older adobe illustrator versions */
+	  else if (!strcmp ((char *)atts[i], "path") || !strcmp((char *)atts[i], "xlink:href"))
+	    href = (const char *)atts[i + 1];
+	}
+    }
+
+  if (!href || x < 0. || y < 0. || w <= 0. || h <= 0.)
+    return;
+  
+  img = gdk_pixbuf_new_from_file (href, NULL);
+  
+  if (!img)
+    return;
+  
+  pixbuf = ctx->pixbuf;
+  has_alpha = gdk_pixbuf_get_has_alpha (img);
+
+  /* composite our source image onto our context's pixbuf 
+     todo: handle current transform affine
+  */
+  gdk_pixbuf_composite (img, pixbuf,
+			x, y, w, h,
+			(double)x, (double)y, 
+			(double)w / (double)gdk_pixbuf_get_width (img),
+			(double)h / (double)gdk_pixbuf_get_height (img),
+			GDK_INTERP_BILINEAR, has_alpha ? 255 : 0);
+  
+  g_object_unref (G_OBJECT (img));
+}
+
+static void
+rsvg_start_line (RsvgHandle *ctx, const xmlChar **atts)
+{
+  int i;
+  double x1 = 0, y1 = 0, x2 = 0, y2 = 0;
+  char * d = NULL;
+  xmlChar *path_atts[3];
+
+  rsvg_parse_style_attrs (ctx, atts);
+  if (atts != NULL)
+    {
+      gint fixed = FALSE;
+      
+      for (i = 0; atts[i] != NULL; i += 2)
+	{
+	  if (!strcmp ((char *)atts[i], "x1"))
+	    x1 = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "y1"))
+	    y1 = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  if (!strcmp ((char *)atts[i], "x2"))
+	    x2 = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "y2"))
+	    y2 = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	}      
+    }
+
+  /* emulate a line using a path */
+  d = g_strdup_printf ("M %f %f L %f %f", x1, y1, x2, y2);
+
+  path_atts[0] = (xmlChar*)"d";
+  path_atts[1] = (xmlChar*)d;
+  path_atts[2] = (xmlChar*)NULL;
+  rsvg_start_path (ctx, (const xmlChar **)path_atts);
+
+  g_free (d);
+}
+
+static void
+rsvg_start_rect (RsvgHandle *ctx, const xmlChar **atts)
+{
+  int i;
+  double x = -1, y = -1, w = -1, h = -1, rx = 0, ry = 0;
+  char * d = NULL;
+  xmlChar *path_atts[3];
+  
+  rsvg_parse_style_attrs (ctx, atts);
+  if (atts != NULL)
+    {
+      gint fixed = FALSE;
+      
+      for (i = 0; atts[i] != NULL; i += 2)
+	{
+	  if (!strcmp ((char *)atts[i], "x"))
+	    x = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "y"))
+	    y = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "width"))
+	    w = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "height"))
+	    h = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "rx"))
+	    rx = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "ry"))
+	    ry = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	}
+    }
+
+  if (x < 0. || y < 0. || w < 0. || h < 0. || rx < 0. || ry < 0.)
+    return;
+
+  /* incrementing y by 1 properly draws borders. this is a HACK */
+  y++;
+
+  /* emulate a rect using a path */
+  d = g_strdup_printf ("M %f %f "
+		       "H %f "
+		       "A %f,%f %f,%f %f %f,%f "
+		       "V %f "
+		       "A %f,%f %f,%f %f %f,%f "
+		       "H %f "
+		       "A %f,%f %f,%f %f %f,%f "
+		       "V %f "
+		       "A %f,%f %f,%f %f %f,%f",
+		       x + rx, y,
+		       x + w - rx,
+		       rx, ry, 0., 0., 1., x + w, y + ry,
+		       y + h - ry,
+		       rx, ry, 0., 0., 1., x + w - rx, y + h,
+		       x + rx,
+		       rx, ry, 0., 0., 1., x, y + h - ry,
+		       y + ry,
+		       rx, ry, 0., 0., 1., x + rx, y);
+
+  path_atts[0] = (xmlChar*)"d";
+  path_atts[1] = (xmlChar*)d;
+  path_atts[2] = (xmlChar*)NULL;
+  rsvg_start_path (ctx, (const xmlChar **)path_atts);
+  
+  g_free (d);
+}
+
+static void
+rsvg_start_circle (RsvgHandle *ctx, const xmlChar **atts)
+{
+  int i;
+  double cx = 0, cy = 0, r = 0;
+  char * d = NULL;
+  xmlChar *path_atts[3];
+  
+  rsvg_parse_style_attrs (ctx, atts);
+  if (atts != NULL)
+    {
+      gint fixed = FALSE;
+      
+      for (i = 0; atts[i] != NULL; i += 2)
+	{
+	  if (!strcmp ((char *)atts[i], "cx"))
+	    cx = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "cy"))
+	    cy = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "r"))
+	    r = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	}
+    }
+
+  if (cx < 0. || cy < 0. || r <= 0.)
+    return;
+
+  /* approximate a circle using 4 bezier paths */
+  d = g_strdup_printf ("M %f %f "
+		       "C %f %f %f %f %f %f "
+		       "C %f %f %f %f %f %f "
+		       "C %f %f %f %f %f %f "
+		       "C %f %f %f %f %f %f "
+		       "Z",
+		       cx + r, cy,
+		       cx + r, cy + r * RSVG_ARC_MAGIC, cx + r * RSVG_ARC_MAGIC, cy + r, cx, cy + r,
+		       cx - r * RSVG_ARC_MAGIC, cy + r, cx - r, cy + r * RSVG_ARC_MAGIC, cx - r, cy,
+		       cx - r, cy - r * RSVG_ARC_MAGIC, cx - r * RSVG_ARC_MAGIC, cy - r, cx, cy - r,
+		       cx + r * RSVG_ARC_MAGIC, cy - r, cx + r, cy - r * RSVG_ARC_MAGIC, cx + r, cy
+		       );
+
+  path_atts[0] = (xmlChar*)"d";
+  path_atts[1] = (xmlChar*)d;
+  path_atts[2] = (xmlChar*)NULL;
+  rsvg_start_path (ctx, (const xmlChar **)path_atts);
+  
+  g_free (d);
+}
+
+static void
+rsvg_start_ellipse (RsvgHandle *ctx, const xmlChar **atts)
+{
+  int i;
+  double cx = 0, cy = 0, rx = 0, ry = 0;
+  char * d = NULL;
+  xmlChar *path_atts[3];
+
+  double r1x, r1y, r2x, r2y, r3x, r3y, r4x, r4y;
+  
+  rsvg_parse_style_attrs (ctx, atts);
+  if (atts != NULL)
+    {
+      gint fixed = FALSE;
+      
+      for (i = 0; atts[i] != NULL; i += 2)
+	{
+	  if (!strcmp ((char *)atts[i], "cx"))
+	    cx = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "cy"))
+	    cy = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "rx"))
+	    rx = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	  else if (!strcmp ((char *)atts[i], "ry"))
+	    ry = rsvg_css_parse_length ((char *)atts[i + 1], RSVG_PIXELS_PER_INCH, &fixed);
+	}
+    }
+
+  if (cx < 0. || cy < 0. || rx <= 0. || ry <= 0.)
+    return;
+
+  r1x = cx - rx ;
+  r1y = cy ;
+  r2x = cx + rx ; 
+  r2y = cy ;
+  r3x = cx ;
+  r3y = cy + ry ;
+  r4x = cx ;
+  r4y = cy - ry ;
+
+  /* approximate an ellipse using 4 bezier paths: todo: this is buggy */
+  d = g_strdup_printf ("M %f %f "
+		       "C %f %f %f %f %f %f "
+		       "C %f %f %f %f %f %f "
+		       "C %f %f %f %f %f %f "
+		       "C %f %f %f %f %f %f "
+		       "Z",
+		       r2x, r2y,
+		       r2x, ( r2y + r4y ) / 2., ( r2x + r4x ) / 2., r4y, r4x, r4y,
+		       ( r4x + r1x ) / 2., r4y, r1x, ( r1y + r4y ) / 2., r1x, r1y,
+		       r1x, ( r1y + r3y ) / 2., ( r1x + r3x ) / 2., r3y, r3x, r3y,
+		       ( r3x + r2x ) / 2., r3y, r2x, ( r3y + r2y ) / 2., r2x, r2y
+		       );
+
+  path_atts[0] = (xmlChar*)"d";
+  path_atts[1] = (xmlChar*)d;
+  path_atts[2] = (xmlChar*)NULL;
+  rsvg_start_path (ctx, (const xmlChar **)path_atts);
+  
+  g_free (d);
+
+  return;
+}
+
 static void
 rsvg_start_element (void *data, const xmlChar *name, const xmlChar **atts)
 {
@@ -1312,8 +1873,22 @@ rsvg_start_element (void *data, const xmlChar *name, const xmlChar **atts)
 	rsvg_start_path (ctx, atts);
       else if (!strcmp ((char *)name, "text"))
 	rsvg_start_text (ctx, atts);
+      else if (!strcmp ((char *)name, "image"))
+	rsvg_start_image (ctx, atts);
+      else if (!strcmp ((char *)name, "line"))
+	rsvg_start_line (ctx, atts);
+      else if (!strcmp ((char *)name, "rect"))
+	rsvg_start_rect (ctx, atts);
+      else if (!strcmp ((char *)name, "circle"))
+	rsvg_start_circle (ctx, atts);
+      else if (!strcmp ((char *)name, "ellipse"))
+	rsvg_start_ellipse (ctx, atts);
       else if (!strcmp ((char *)name, "defs"))
 	rsvg_start_defs (ctx, atts);
+      else if (!strcmp ((char *)name, "polygon"))
+	rsvg_start_polygon (ctx, atts);
+      else if (!strcmp ((char *)name, "polyline"))
+	rsvg_start_polyline (ctx, atts);
       else if (!strcmp ((char *)name, "linearGradient"))
 	rsvg_start_linear_gradient (ctx, atts);
       else if (!strcmp ((char *)name, "radialGradient"))
@@ -1566,8 +2141,13 @@ rsvg_handle_close (RsvgHandle  *handle,
   GError *real_error;
 
   handle->error = &real_error;
-  xmlParseChunk (handle->ctxt, chars, 1, TRUE);
-  xmlFreeParserCtxt (handle->ctxt);
+
+  if (handle->ctxt != NULL)
+  {
+    xmlParseChunk (handle->ctxt, chars, 1, TRUE);
+    xmlFreeParserCtxt (handle->ctxt);
+  }
+  
   /* FIXME: Error handling not implemented. */
   /*
   if (real_error != NULL)
@@ -1793,7 +2373,7 @@ rsvg_pixbuf_from_file_at_zoom (const gchar *file_name,
  * Loads a new #GdkPixbuf from @file_name and returns it.  This pixbuf is scaled
  * from the size indicated by the file by a factor of @x_zoom and @y_zoom. If the
  * resulting pixbuf would be larger than max_width/max_heigh it is uniformly scaled
- * down to fit in that rectangle.The caller must assume the reference to the
+ * down to fit in that rectangle. The caller must assume the reference to the
  * returned pixbuf. If an error occurred, @error is set and %NULL is returned.
  * 
  * Return value: A newly allocated #GdkPixbuf, or %NULL
@@ -1944,5 +2524,3 @@ rsvg_pixbuf_from_file_at_max_size (const gchar     *file_name,
   rsvg_handle_free (handle);
   return retval;
 }
-
-
